@@ -1,10 +1,10 @@
 /* -*- c-basic-offset: 2 -*- */
 /* gifsicle.c - gifsicle's main loop.
-   Copyright (C) 1997-2011 Eddie Kohler, ekohler@cs.ucla.edu
+   Copyright (C) 1997-2013 Eddie Kohler, ekohler@cs.ucla.edu
    This file is part of gifsicle.
 
    Gifsicle is free software. It is distributed under the GNU Public License,
-   version 2 or later; you can copy, distribute, or alter it at will, as long
+   version 2; you can copy, distribute, or alter it at will, as long
    as this notice is kept intact and this source code is made available. There
    is no warranty, express or implied. */
 
@@ -37,7 +37,7 @@ static int unoptimizing = 0;
 
 static int gif_read_flags = 0;
 static int nextfile = 0;
-int gif_write_flags = 0;
+Gif_CompressInfo gif_write_info;
 
 static int frames_done = 0;
 static int files_given = 0;
@@ -91,7 +91,7 @@ static const char *input_option_types[] = {
 
 /* output option types */
 static Gt_OutputData def_output_data;
-static Gt_OutputData active_output_data;
+Gt_OutputData active_output_data;
 static int next_output = 0;
 static int active_next_output = 0;
 static int any_output_successful = 0;
@@ -179,6 +179,7 @@ static const char *output_option_types[] = {
 #define RESIZE_FIT_OPT		364
 #define RESIZE_FIT_WIDTH_OPT	365
 #define RESIZE_FIT_HEIGHT_OPT	366
+#define SIZE_INFO_OPT		367
 
 #define LOOP_TYPE		(Clp_ValFirstUser)
 #define DISPOSAL_TYPE		(Clp_ValFirstUser + 1)
@@ -269,8 +270,6 @@ const Clp_Option options[] = {
   { "rotate-270", 0, ROTATE_270_OPT, 0, 0 },
   { "no-rotate", 0, NO_ROTATE_OPT, 0, 0 },
 
-  { "scale", 0, SCALE_OPT, SCALE_FACTOR_TYPE, Clp_Negate },
-  { "screen", 0, LOGICAL_SCREEN_OPT, DIMENSIONS_TYPE, Clp_Negate },
   { "same-background", 0, SAME_BACKGROUND_OPT, 0, 0 },
   { "same-bg", 0, SAME_BACKGROUND_OPT, 0, 0 },
   { "same-clip", 0, SAME_CROP_OPT, 0, 0 },
@@ -286,6 +285,10 @@ const Clp_Option options[] = {
   { "same-position", 0, SAME_POSITION_OPT, 0, 0 },
   { "same-screen", 0, SAME_LOGICAL_SCREEN_OPT, 0, 0 },
   { "same-transparent", 0, SAME_TRANSPARENT_OPT, 0, 0 },
+  { "scale", 0, SCALE_OPT, SCALE_FACTOR_TYPE, Clp_Negate },
+  { "screen", 0, LOGICAL_SCREEN_OPT, DIMENSIONS_TYPE, Clp_Negate },
+  { "sinfo", 0, SIZE_INFO_OPT, 0, Clp_Negate },
+  { "size-info", 0, SIZE_INFO_OPT, 0, Clp_Negate },
 
   { "transform-colormap", 0, COLOR_TRANSFORM_OPT, Clp_ValStringNotOption,
     Clp_Negate },
@@ -314,13 +317,18 @@ static void redundant_option_warning(const char *);
 static void
 set_mode(int newmode)
 {
-  if (mode == BLANK_MODE)
-    mode = newmode;
-  else if (mode == newmode)
-    /* do nothing */;
-  else if (newmode == INFOING)
+  if (mode == BLANK_MODE) {
+    if (newmode != BLANK_MODE)
+      mode = newmode;
+    else if (infoing == 1)
+      mode = INFOING;
+    else
+      mode = MERGING;
+  }
+
+  if (mode != INFOING && infoing == 1)
     fatal_error("'--info' suppresses normal output, can't use with an\n  output mode like '--merge' or '--batch'.\n  (Try '-II', which doesn't suppress normal output.)");
-  else
+  if (newmode != BLANK_MODE && newmode != mode)
     fatal_error("too late to change modes");
 }
 
@@ -331,8 +339,7 @@ set_frame_change(int kind)
   int i;
   Gt_Frameset *fset;
 
-  if (mode == BLANK_MODE)
-    set_mode(infoing == 1 ? INFOING : MERGING);
+  set_mode(BLANK_MODE);
   if (mode < DELETING && frames_done) {
     fatal_error("frame selection and frame changes don't mix");
     return;
@@ -378,7 +385,7 @@ set_frame_change(int kind)
   }
 }
 
-void
+static void
 frame_change_done(void)
 {
   if (nested_mode)
@@ -390,7 +397,7 @@ frame_change_done(void)
 }
 
 
-void
+static void
 show_frame(int imagenumber, int usename)
 {
   Gif_Image *gfi;
@@ -405,6 +412,7 @@ show_frame(int imagenumber, int usename)
    case INSERTING:
    case EXPLODING:
    case INFOING:
+   case BATCHING:
     if (!frames_done)
       clear_frameset(frames, first_input_frame);
     frame = add_frame(frames, -1, input, gfi);
@@ -412,10 +420,6 @@ show_frame(int imagenumber, int usename)
       frame->explode_by_name = 1;
     break;
 
-   case BATCHING:
-    add_frame(frames, first_input_frame + imagenumber, input, gfi);
-    break;
-
    case DELETING:
     frame = &FRAME(frames, first_input_frame + imagenumber);
     frame->use = 0;
@@ -583,8 +587,7 @@ input_stream(const char *name)
     combine_output_options();
   files_given++;
 
-  if (infoing == 1)
-    set_mode(INFOING);
+  set_mode(BLANK_MODE);
 
   f = open_giffile(name);
   if (!f)
@@ -648,8 +651,7 @@ input_stream(const char *name)
   input = gfs;
 
   /* Processing when we've got a new input frame */
-  if (mode == BLANK_MODE)
-    set_mode(MERGING);
+  set_mode(BLANK_MODE);
 
   if (active_output_data.output_name == 0) {
     /* Don't override explicit output names.
@@ -871,7 +873,7 @@ write_stream(const char *output_name, Gif_Stream *gfs)
   }
 
   if (f) {
-    Gif_FullWriteFile(gfs, gif_write_flags, f);
+    Gif_FullWriteFile(gfs, &gif_write_info, f);
     fclose(f);
     any_output_successful = 1;
   } else
@@ -888,6 +890,7 @@ merge_and_write_frames(const char *outfile, int f1, int f2)
   assert(!nested_mode);
   if (verbosing)
     verbose_open('[', outfile ? outfile : "#stdout#");
+  active_output_data.active_output_name = outfile;
 
   colormap_change = active_output_data.colormap_size > 0
     || active_output_data.colormap_fixed;
@@ -924,6 +927,7 @@ merge_and_write_frames(const char *outfile, int f1, int f2)
   }
 
   if (verbosing) verbose_close(']');
+  active_output_data.active_output_name = 0;
 }
 
 static void
@@ -954,12 +958,11 @@ output_information(const char *outfile)
       fr = &FRAME(frames, i);
       gfs = fr->stream;
       gfs->userflags = 0;
-      stream_info(f, gfs, fr->input_filename, fr->colormap_info,
-		  fr->extensions_info);
+      stream_info(f, gfs, fr->input_filename, fr->info_flags);
       for (j = i; j < frames->count; j++)
 	if (FRAME(frames, j).stream == gfs) {
 	  fr = &FRAME(frames, j);
-	  image_info(f, gfs, fr->image, fr->colormap_info);
+	  image_info(f, gfs, fr->image, fr->info_flags);
 	}
     }
 
@@ -1292,6 +1295,7 @@ main(int argc, char *argv[])
 
   frames = new_frameset(16);
   initialize_def_frame();
+  Gif_InitCompressInfo(&gif_write_info);
 
 #ifdef DMALLOC
   dmalloc_verbose("fudge");
@@ -1345,9 +1349,9 @@ main(int argc, char *argv[])
 
      case COLOR_INFO_OPT:
       if (clp->negated)
-	def_frame.colormap_info = 0;
+	def_frame.info_flags &= ~INFO_COLORMAPS;
       else {
-	def_frame.colormap_info = 1;
+	def_frame.info_flags |= INFO_COLORMAPS;
 	if (!infoing)
 	  infoing = 1;
       }
@@ -1355,9 +1359,19 @@ main(int argc, char *argv[])
 
      case EXTENSION_INFO_OPT:
       if (clp->negated)
-	def_frame.extensions_info = 0;
+	def_frame.info_flags &= ~INFO_EXTENSIONS;
       else {
-	def_frame.extensions_info = 1;
+	def_frame.info_flags |= INFO_EXTENSIONS;
+	if (!infoing)
+	  infoing = 1;
+      }
+      break;
+
+     case SIZE_INFO_OPT:
+      if (clp->negated)
+	def_frame.info_flags &= ~INFO_SIZES;
+      else {
+	def_frame.info_flags |= INFO_SIZES;
 	if (!infoing)
 	  infoing = 1;
       }
@@ -1635,9 +1649,9 @@ main(int argc, char *argv[])
 
      case CAREFUL_OPT: {
        if (clp->negated)
-	 gif_write_flags = 0;
+	 gif_write_info.flags = 0;
        else
-	 gif_write_flags = GIF_WRITE_CAREFUL_MIN_CODE_SIZE | GIF_WRITE_EAGER_CLEAR;
+	 gif_write_info.flags = GIF_WRITE_CAREFUL_MIN_CODE_SIZE | GIF_WRITE_EAGER_CLEAR;
        break;
      }
 
@@ -1787,7 +1801,7 @@ main(int argc, char *argv[])
 #else
       printf("LCDF Gifsicle %s\n", VERSION);
 #endif
-      printf("Copyright (C) 1997-2011 Eddie Kohler\n\
+      printf("Copyright (C) 1997-2013 Eddie Kohler\n\
 This is free software; see the source for copying conditions.\n\
 There is NO warranty, not even for merchantability or fitness for a\n\
 particular purpose.\n");
