@@ -3,7 +3,7 @@
    This file is part of gifsicle.
 
    Gifsicle is free software. It is distributed under the GNU Public License,
-   version 2 or later; you can copy, distribute, or alter it at will, as long
+   version 2; you can copy, distribute, or alter it at will, as long
    as this notice is kept intact and this source code is made available. There
    is no warranty, express or implied. */
 
@@ -48,8 +48,7 @@ typedef struct Gt_Frame {
 
   unsigned flip_horizontal: 1;
   unsigned flip_vertical: 1;
-  unsigned colormap_info: 1;
-  unsigned extensions_info: 1;
+  unsigned info_flags: 3;
   unsigned position_is_offset: 1;
   unsigned total_crop: 1;
   unsigned rotation;
@@ -95,6 +94,7 @@ struct Gt_ColorTransform {
 typedef struct {
 
   const char *output_name;
+  const char *active_output_name;
 
   int screen_width;
   int screen_height;
@@ -119,6 +119,8 @@ typedef struct {
 
 } Gt_OutputData;
 
+extern Gt_OutputData active_output_data;
+
 #define GT_SCALING_NONE		0
 #define GT_SCALING_RESIZE	1
 #define GT_SCALING_SCALE	2
@@ -146,7 +148,7 @@ extern const char *program_name;
 extern int verbosing;
 extern int error_count;
 extern int no_warnings;
-extern int gif_write_flags;
+extern Gif_CompressInfo gif_write_info;
 
 void fatal_error(const char *message, ...) NORETURN;
 void warning(int need_file, const char *message, ...);
@@ -167,9 +169,11 @@ void verbose_endline(void);
 /*****
  * info &c
  **/
-void stream_info(FILE *, Gif_Stream *, const char *,
-		 int colormaps, int extensions);
-void image_info(FILE *, Gif_Stream *, Gif_Image *, int colormaps);
+#define INFO_COLORMAPS	1
+#define INFO_EXTENSIONS	2
+#define INFO_SIZES	4
+void stream_info(FILE *f, Gif_Stream *gfs, const char *filename, int flags);
+void image_info(FILE *f, Gif_Stream *gfs, Gif_Image *gfi, int flags);
 
 char *explode_filename(const char *filename, int number,
 		       const char *name, int max_nimg);
@@ -179,6 +183,8 @@ char *explode_filename(const char *filename, int number,
  **/
 void	unmark_colors(Gif_Colormap *);
 void	unmark_colors_2(Gif_Colormap *);
+void	mark_used_colors(Gif_Stream *gfs, Gif_Image *gfi, Gt_Crop *crop,
+                         int compress_immediately);
 int	find_color_index(Gif_Color *c, int nc, Gif_Color *);
 int	merge_colormap_if_possible(Gif_Colormap *, Gif_Colormap *);
 
@@ -208,6 +214,7 @@ Gt_ColorTransform *append_color_change
 
 void	pipe_color_transformer(Gif_Colormap *, void *);
 
+void	combine_crop(Gt_Crop *dstcrop, const Gt_Crop *srccrop, const Gif_Image *gfi);
 int	crop_image(Gif_Image *, Gt_Crop *, int preserve_total_crop);
 
 void	flip_image(Gif_Image *, int scr_width, int scr_height, int is_vert);
