@@ -1,6 +1,6 @@
 /* -*- mode: c; c-basic-offset: 2 -*- */
 /* gifwrite.c - Functions to write GIFs.
-   Copyright (C) 1997-2011 Eddie Kohler, kohler@cs.ucla.edu
+   Copyright (C) 1997-2013 Eddie Kohler, kohler@cs.ucla.edu
    This file is part of the LCDF GIF library.
 
    The LCDF GIF library is free software. It is distributed under the GNU
@@ -17,6 +17,8 @@
 #include <lcdfgif/gif.h>
 #include <stdarg.h>
 #include <string.h>
+#include <assert.h>
+#include <limits.h>
 #ifdef __cplusplus
 extern "C" {
 #endif
@@ -48,7 +50,6 @@ extern "C" {
 #define LINKS_TYPE		1
 #define MAX_LINKS_TYPE		5
 typedef struct Gif_Node {
-
   Gif_Code code;
   uint8_t type;
   uint8_t suffix;
@@ -57,34 +58,30 @@ typedef struct Gif_Node {
     struct Gif_Node *s;
     struct Gif_Node **m;
   } child;
-
 } Gif_Node;
 
 
-typedef struct Gif_Context {
-
+typedef struct Gif_CodeTable {
   Gif_Node *nodes;
   int nodes_pos;
   Gif_Node **links;
   int links_pos;
-
-} Gif_Context;
+  int clear_code;
+} Gif_CodeTable;
 
 
 typedef struct Gif_Writer {
-
   FILE *f;
   uint8_t *v;
   uint32_t pos;
   uint32_t cap;
-  int flags;
+  Gif_CompressInfo gcinfo;
   int global_size;
   int local_size;
   int errors;
   int cleared;
   void (*byte_putter)(uint8_t, struct Gif_Writer *);
-  void (*block_putter)(uint8_t *, uint16_t, struct Gif_Writer *);
-
+  void (*block_putter)(const uint8_t *, uint16_t, struct Gif_Writer *);
 } Gif_Writer;
 
 
@@ -106,9 +103,9 @@ file_byte_putter(uint8_t b, Gif_Writer *grr)
 }
 
 static void
-file_block_putter(uint8_t *block, uint16_t size, Gif_Writer *grr)
+file_block_putter(const uint8_t *block, uint16_t size, Gif_Writer *grr)
 {
-  if (fwrite(block, 1, size, grr->f) != size)
+  if (fwrite(block, 1, size, grr->f) != (size_t) size)
     grr->errors = 1;
 }
 
@@ -117,7 +114,7 @@ static void
 memory_byte_putter(uint8_t b, Gif_Writer *grr)
 {
   if (grr->pos >= grr->cap) {
-    grr->cap *= 2;
+    grr->cap = (grr->cap ? grr->cap * 2 : 1024);
     Gif_ReArray(grr->v, uint8_t, grr->cap);
   }
   if (grr->v) {
@@ -127,10 +124,10 @@ memory_byte_putter(uint8_t b, Gif_Writer *grr)
 }
 
 static void
-memory_block_putter(uint8_t *data, uint16_t len, Gif_Writer *grr)
+memory_block_putter(const uint8_t *data, uint16_t len, Gif_Writer *grr)
 {
-  if (grr->pos + len >= grr->cap) {
-    grr->cap *= 2;
+  while (grr->pos + len >= grr->cap) {
+    grr->cap = (grr->cap ? grr->cap * 2 : 1024);
     Gif_ReArray(grr->v, uint8_t, grr->cap);
   }
   if (grr->v) {
@@ -140,17 +137,58 @@ memory_block_putter(uint8_t *data, uint16_t len, Gif_Writer *grr)
 }
 
 
+static int
+gfc_init(Gif_CodeTable *gfc)
+{
+  gfc->nodes = Gif_NewArray(Gif_Node, NODES_SIZE);
+  gfc->links = Gif_NewArray(Gif_Node *, LINKS_SIZE);
+  return gfc->nodes && gfc->links;
+}
+
+static inline void
+gfc_clear(Gif_CodeTable *gfc, Gif_Code clear_code)
+{
+  int c;
+  /* The first clear_code nodes are reserved for single-pixel codes */
+  gfc->nodes_pos = clear_code;
+  gfc->links_pos = 0;
+  for (c = 0; c < clear_code; c++) {
+    gfc->nodes[c].code = c;
+    gfc->nodes[c].type = LINKS_TYPE;
+    gfc->nodes[c].suffix = c;
+    gfc->nodes[c].child.s = 0;
+  }
+  gfc->clear_code = clear_code;
+}
+
+static inline Gif_Node *
+gfc_lookup(Gif_CodeTable *gfc, Gif_Node *node, uint8_t suffix)
+{
+  assert(!node || (node >= gfc->nodes && node < gfc->nodes + NODES_SIZE));
+  assert(suffix < gfc->clear_code);
+  if (!node)
+    return &gfc->nodes[suffix];
+  else if (node->type == TABLE_TYPE)
+    return node->child.m[suffix];
+  else {
+    for (node = node->child.s; node; node = node->sibling)
+      if (node->suffix == suffix)
+        return node;
+    return NULL;
+  }
+}
+
 static void
-change_node_to_table(Gif_Context *gfc, Gif_Node *work_node,
-		     Gif_Node *next_node, Gif_Code clear_code)
+gfc_change_node_to_table(Gif_CodeTable *gfc, Gif_Node *work_node,
+                         Gif_Node *next_node)
 {
   /* change links node to table node */
   Gif_Code c;
   Gif_Node **table = &gfc->links[gfc->links_pos];
   Gif_Node *n;
-  gfc->links_pos += clear_code;
+  gfc->links_pos += gfc->clear_code;
 
-  for (c = 0; c < clear_code; c++)
+  for (c = 0; c < gfc->clear_code; c++)
     table[c] = 0;
   table[next_node->suffix] = next_node;
   for (n = work_node->child.s; n; n = n->sibling)
@@ -160,110 +198,149 @@ change_node_to_table(Gif_Context *gfc, Gif_Node *work_node,
   work_node->child.m = table;
 }
 
-
-static void
-write_compressed_data(uint8_t **img, unsigned width, unsigned height,
-		      int min_code_bits, Gif_Context *gfc, Gif_Writer *grr)
+static inline void
+gfc_define(Gif_CodeTable *gfc, Gif_Node *work_node, uint8_t suffix,
+           Gif_Code next_code)
 {
-  uint8_t buffer[WRITE_BUFFER_SIZE];
-  uint8_t *buf;
+  /* Add a new code to our dictionary. First reserve a node for the
+     added code. It's LINKS_TYPE at first. */
+  Gif_Node *next_node = &gfc->nodes[gfc->nodes_pos];
+  gfc->nodes_pos++;
+  next_node->code = next_code;
+  next_node->type = LINKS_TYPE;
+  next_node->suffix = suffix;
+  next_node->child.s = 0;
 
-  unsigned xleft;
-  uint8_t *imageline;
+  /* link next_node into work_node's set of children */
+  if (work_node->type == TABLE_TYPE)
+    work_node->child.m[suffix] = next_node;
+  else if (work_node->type < MAX_LINKS_TYPE
+           || gfc->links_pos + gfc->clear_code > LINKS_SIZE) {
+    next_node->sibling = work_node->child.s;
+    work_node->child.s = next_node;
+    if (work_node->type < MAX_LINKS_TYPE)
+      work_node->type++;
+  } else
+    gfc_change_node_to_table(gfc, work_node, next_node);
+}
 
-  unsigned leftover;
-  int bits_left_over;
+static inline const uint8_t *
+gif_imageline(Gif_Image *gfi, unsigned pos)
+{
+  unsigned y = pos / gfi->width, x = pos - y * gfi->width;
+  if (y == (unsigned) gfi->height)
+    return NULL;
+  else if (!gfi->interlace)
+    return gfi->img[y] + x;
+  else
+    return gfi->img[Gif_InterlaceLine(y, gfi->height)] + x;
+}
+
+static inline unsigned
+gif_line_endpos(Gif_Image *gfi, unsigned pos)
+{
+  unsigned y = pos / gfi->width;
+  return (y + 1) * gfi->width;
+}
+
+static int
+write_compressed_data(Gif_Image *gfi,
+		      int min_code_bits, Gif_CodeTable *gfc, Gif_Writer *grr)
+{
+  uint8_t stack_buffer[232];
+  uint8_t *buf = stack_buffer;
+  unsigned bufpos = 0;
+  unsigned bufcap = sizeof(stack_buffer) * 8;
+
+  unsigned pos;
+  unsigned clear_bufpos, clear_pos;
+  unsigned line_endpos;
+  const uint8_t *imageline;
 
   Gif_Node *work_node;
-  int work_depth;
+  unsigned run;
+#define RUN_EWMA_SHIFT 4
+#define RUN_EWMA_SCALE 19
+#define RUN_INV_THRESH ((unsigned) (1 << RUN_EWMA_SCALE) / 3000)
+  unsigned run_ewma;
   Gif_Node *next_node;
   Gif_Code next_code = 0;
   Gif_Code output_code;
-  Gif_Code clear_code;
-  Gif_Code eoi_code;
 #define CUR_BUMP_CODE (1 << cur_code_bits)
   uint8_t suffix;
 
-  int end_table_avg_depth = 0;
-  int end_table_count = 0;
-
   int cur_code_bits;
 
   /* Here we go! */
   gifputbyte(min_code_bits, grr);
-  clear_code = 1 << min_code_bits;
-  eoi_code = clear_code + 1;
+#define CLEAR_CODE	((Gif_Code) (1 << min_code_bits))
+#define EOI_CODE	((Gif_Code) (CLEAR_CODE + 1))
   grr->cleared = 0;
 
   cur_code_bits = min_code_bits + 1;
   /* next_code set by first runthrough of output clear_code */
-  GIF_DEBUG(("clear(%d) eoi(%d) bits(%d)",clear_code,eoi_code,cur_code_bits));
+  GIF_DEBUG(("clear(%d) eoi(%d) bits(%d)", CLEAR_CODE, EOI_CODE, cur_code_bits));
 
   work_node = 0;
-  work_depth = 0;
-  output_code = clear_code;
+  run = 0;
+  run_ewma = 1 << RUN_EWMA_SCALE;
+  output_code = CLEAR_CODE;
   /* Because output_code is clear_code, we'll initialize next_code, et al.
      below. */
 
-  bits_left_over = 0;
-  leftover = 0;
-  buf = buffer;
-  xleft = width;
-  imageline = img[0];
+  pos = clear_pos = clear_bufpos = 0;
+  line_endpos = gfi->width;
+  imageline = gif_imageline(gfi, pos);
 
   while (1) {
 
     /*****
-     * Output 'output_code' to the data stream. */
-
-    leftover |= output_code << bits_left_over;
-    bits_left_over += cur_code_bits;
-    while (bits_left_over >= 8) {
-      *buf++ = leftover & 0xFF;
-      leftover = (leftover >> 8) & 0x00FFFFFF;
-      bits_left_over -= 8;
-      if (buf == buffer + WRITE_BUFFER_SIZE) {
-	gifputbyte(WRITE_BUFFER_SIZE, grr);
-	gifputblock(buffer, WRITE_BUFFER_SIZE, grr);
-	buf = buffer;
-      }
+     * Output 'output_code' to the memory buffer. */
+    if (bufpos + cur_code_bits >= bufcap) {
+      unsigned ncap = bufcap * 2 + (24 << 3);
+      uint8_t *nbuf = Gif_NewArray(uint8_t, ncap >> 3);
+      if (!nbuf)
+        return 0;
+      memcpy(nbuf, buf, bufcap >> 3);
+      if (buf != stack_buffer)
+        Gif_DeleteArray(buf);
+      buf = nbuf;
+      bufcap = ncap;
     }
 
-    if (output_code == clear_code) {
+    {
+      unsigned startpos = bufpos;
+      do {
+        if (bufpos & 7)
+          buf[bufpos >> 3] |= output_code << (bufpos & 7);
+        else
+          buf[bufpos >> 3] = output_code >> (bufpos - startpos);
+
+        bufpos += 8 - (bufpos & 7);
+      } while (bufpos < startpos + cur_code_bits);
+      bufpos = startpos + cur_code_bits;
+    }
+
+
+    /*****
+     * Handle special codes. */
+
+    if (output_code == CLEAR_CODE) {
       /* Clear data and prepare gfc */
-      Gif_Code c;
-
       cur_code_bits = min_code_bits + 1;
-      next_code = eoi_code + 1;
+      next_code = EOI_CODE + 1;
+      run_ewma = 1 << RUN_EWMA_SCALE;
+      gfc_clear(gfc, CLEAR_CODE);
+      clear_pos = clear_bufpos = 0;
 
-      /* The first clear_code nodes are reserved for single-pixel codes */
-      gfc->nodes_pos = clear_code;
-      gfc->links_pos = 0;
-      for (c = 0; c < clear_code; c++) {
-	gfc->nodes[c].code = c;
-	gfc->nodes[c].type = LINKS_TYPE;
-	gfc->nodes[c].suffix = c;
-	gfc->nodes[c].child.s = 0;
-      }
+      GIF_DEBUG(("clear"));
 
-      end_table_count = 0;
-      end_table_avg_depth = 0;
-
-    } else if (output_code == eoi_code)
+    } else if (output_code == EOI_CODE)
       break;
 
-    else if (next_code > CUR_BUMP_CODE) {
+    else if (next_code > CUR_BUMP_CODE && cur_code_bits < GIF_MAX_CODE_BITS)
       /* bump up compression size */
-      if (cur_code_bits < GIF_MAX_CODE_BITS)
-	++cur_code_bits;
-      else if ((end_table_count > 8 /* totally arbitrary constants */
-		&& end_table_avg_depth < 12 * 128)
-	       || (grr->flags & GIF_WRITE_EAGER_CLEAR)) {
-	output_code = clear_code;
-	grr->cleared = 1;
-	continue;
-      }
-    }
+      ++cur_code_bits;
 
 
     /*****
@@ -271,109 +348,122 @@ write_compressed_data(uint8_t **img, unsigned width, unsigned height,
 
     /* If height is 0 -- no more pixels to write -- we output work_node next
        time around. */
-    while (height != 0) {
+    while (imageline) {
       suffix = *imageline;
-      if (suffix >= clear_code)
-	/* should not happen unless GIF_WRITE_CAREFUL_MIN_CODE_BITS */
-	suffix = 0;
-      if (!work_node)
-	next_node = &gfc->nodes[suffix];
-      else if (work_node->type == TABLE_TYPE)
-	next_node = work_node->child.m[suffix];
-      else
-	for (next_node = work_node->child.s; next_node;
-	     next_node = next_node->sibling)
-	  if (next_node->suffix == suffix)
-	    break;
+      next_node = gfc_lookup(gfc, work_node, suffix);
 
       imageline++;
-      xleft--;
-      if (xleft == 0) {
-	xleft = width;
-	height--;
-	img++;
-	imageline = img[0];
+      pos++;
+      if (pos == line_endpos) {
+	imageline = gif_imageline(gfi, pos);
+        line_endpos += gfi->width;
       }
 
       if (!next_node) {
 	/* Output the current code. */
 	if (next_code < GIF_MAX_CODE) {
-	  /* Add a new code to our dictionary. First reserve a node for the
-	     added code. It's LINKS_TYPE at first. */
-	  next_node = &gfc->nodes[gfc->nodes_pos];
-	  gfc->nodes_pos++;
-	  next_node->code = next_code;
-	  next_code++;
-	  next_node->type = LINKS_TYPE;
-	  next_node->suffix = suffix;
-	  next_node->child.s = 0;
-
-	  /* link next_node into work_node's set of children */
-	  if (work_node->type == TABLE_TYPE)
-	    work_node->child.m[suffix] = next_node;
-	  else if (work_node->type < MAX_LINKS_TYPE
-		   || gfc->links_pos + clear_code > LINKS_SIZE) {
-	    next_node->sibling = work_node->child.s;
-	    work_node->child.s = next_node;
-	    work_node->type++;
-	  } else
-	    change_node_to_table(gfc, work_node, next_node, clear_code);
-	} else {
+          gfc_define(gfc, work_node, suffix, next_code);
+          next_code++;
+	} else
 	  next_code = GIF_MAX_CODE + 1; /* to match "> CUR_BUMP_CODE" above */
-	  if (end_table_count < 256)
-	    ++end_table_count;
-	  end_table_avg_depth +=
-	    (work_depth * 128 - end_table_avg_depth) / end_table_count;
-	}
+
+        /* Check whether to clear table. */
+        if (next_code > 4094) {
+          int do_clear = grr->gcinfo.flags & GIF_WRITE_EAGER_CLEAR;
+
+          if (!do_clear) {
+            unsigned pixels_left = gfi->width * gfi->height - pos;
+            if (pixels_left) {
+              /* Always clear if run_ewma gets small relative to
+                 min_code_bits. Otherwise, clear if #images/run is smaller
+                 than an empirical threshold, meaning it will take more than
+                 3000 or so average runs to complete the image. */
+              if (run_ewma < ((36U << RUN_EWMA_SCALE) / min_code_bits)
+                  || pixels_left > UINT_MAX / RUN_INV_THRESH
+                  || run_ewma < pixels_left * RUN_INV_THRESH)
+                do_clear = 1;
+            }
+          }
+
+          if ((do_clear || run < 7) && !clear_pos) {
+            clear_pos = pos - (run + 1);
+            clear_bufpos = bufpos;
+          } else if (!do_clear && run > 50)
+            clear_pos = clear_bufpos = 0;
+
+          if (do_clear) {
+            GIF_DEBUG(("rewind %u pixels/%d bits", pos - clear_pos, bufpos + cur_code_bits - clear_bufpos));
+            output_code = CLEAR_CODE;
+            pos = clear_pos;
+            imageline = gif_imageline(gfi, pos);
+            line_endpos = gif_line_endpos(gfi, pos);
+            bufpos = clear_bufpos;
+            buf[bufpos >> 3] &= (1 << (bufpos & 7)) - 1;
+            work_node = 0;
+            run = 0;
+            grr->cleared = 1;
+            goto found_output_code;
+          }
+        }
+
+        /* Adjust current run length average. */
+        run = (run << RUN_EWMA_SCALE) + (1 << (RUN_EWMA_SHIFT - 1));
+        if (run < run_ewma)
+          run_ewma -= (run_ewma - run) >> RUN_EWMA_SHIFT;
+        else
+          run_ewma += (run - run_ewma) >> RUN_EWMA_SHIFT;
 
 	/* Output the current code. */
 	output_code = work_node->code;
 	work_node = &gfc->nodes[suffix];
-	work_depth = 1;
+	run = 1;
 	goto found_output_code;
       }
 
       work_node = next_node;
-      ++work_depth;
+      ++run;
     }
 
     /* Ran out of data if we get here. */
-    output_code = (work_node ? work_node->code : eoi_code);
+    output_code = (work_node ? work_node->code : EOI_CODE);
     work_node = 0;
-    work_depth = 0;
+    run = 0;
 
    found_output_code: ;
   }
 
-  if (bits_left_over > 0)
-    *buf++ = leftover;
-
-  if (buf != buffer) {
-    GIF_DEBUG(("imageblock(%d)", buf - buffer));
-    gifputbyte(buf - buffer, grr);
-    gifputblock(buffer, buf - buffer, grr);
+  /* Output memory buffer to stream. */
+  {
+    unsigned outpos = 0;
+    bufpos = (bufpos + 7) >> 3;
+    while (outpos < bufpos) {
+      unsigned w = (bufpos - outpos > 255 ? 255 : bufpos - outpos);
+      gifputbyte(w, grr);
+      gifputblock(buf + outpos, w, grr);
+      outpos += w;
+    }
+    gifputbyte(0, grr);
   }
 
-  gifputbyte(0, grr);
+  if (buf != stack_buffer)
+    Gif_DeleteArray(buf);
+
+  return 1;
 }
 
 
 static int
-calculate_min_code_bits(Gif_Stream *gfs, Gif_Image *gfi, Gif_Writer *grr)
+calculate_min_code_bits(Gif_Image *gfi, const Gif_Writer *grr)
 {
   int colors_used = -1, min_code_bits, i;
 
-  if (grr->flags & GIF_WRITE_CAREFUL_MIN_CODE_SIZE) {
+  if (grr->gcinfo.flags & GIF_WRITE_CAREFUL_MIN_CODE_SIZE) {
     /* calculate m_c_b based on colormap */
     if (grr->local_size > 0)
       colors_used = grr->local_size;
     else if (grr->global_size > 0)
       colors_used = grr->global_size;
 
-  } else if (gfi->compressed) {
-    /* take m_c_b from compressed image */
-    colors_used = 1 << gfi->compressed[0];
-
   } else if (gfi->img) {
     /* calculate m_c_b from uncompressed data */
     int x, y, width = gfi->width, height = gfi->height;
@@ -386,6 +476,10 @@ calculate_min_code_bits(Gif_Stream *gfs, Gif_Image *gfi, Gif_Writer *grr)
     }
     colors_used++;
 
+  } else if (gfi->compressed) {
+    /* take m_c_b from compressed image */
+    colors_used = 1 << gfi->compressed[0];
+
   } else {
     /* should never happen */
     colors_used = 256;
@@ -398,101 +492,75 @@ calculate_min_code_bits(Gif_Stream *gfs, Gif_Image *gfi, Gif_Writer *grr)
     i *= 2;
   }
 
-  if ((grr->flags & GIF_WRITE_CAREFUL_MIN_CODE_SIZE)
-      && gfi->compressed && gfi->compressed[0] != min_code_bits) {
-    /* if compressed image disagrees with careful min_code_bits, recompress */
-    if (Gif_UncompressImage(gfi))
-      Gif_FullCompressImage(gfs, gfi, grr->flags);
-  }
-
   return min_code_bits;
 }
 
-static int
-write_image_data(Gif_Image *gfi, uint8_t min_code_bits,
-		 Gif_Context *gfc, Gif_Writer *grr)
+
+static int get_color_table_size(const Gif_Stream *gfs, Gif_Image *gfi,
+				Gif_Writer *grr);
+
+static void
+save_compression_result(Gif_Image *gfi, Gif_Writer *grr, int ok)
 {
-  uint8_t **img = gfi->img;
-  uint16_t width = gfi->width, height = gfi->height;
-
-  if (gfi->interlace) {
-    uint16_t y;
-    uint8_t **nimg = Gif_NewArray(uint8_t *, height + 1);
-    if (!nimg) return 0;
-
-    for (y = 0; y < height; y++)
-      nimg[y] = img[Gif_InterlaceLine(y, height)];
-    nimg[height] = 0;
-
-    write_compressed_data(nimg, width, height, min_code_bits, gfc, grr);
-
-    Gif_DeleteArray(nimg);
-  } else
-    write_compressed_data(img, width, height, min_code_bits, gfc, grr);
-
-  return 1;
+  if (!(grr->gcinfo.flags & GIF_WRITE_SHRINK)
+      || (ok && (!gfi->compressed || gfi->compressed_len > grr->pos))) {
+    if (gfi->compressed)
+      (*gfi->free_compressed)((void *) gfi->compressed);
+    if (ok) {
+      gfi->compressed = grr->v;
+      gfi->compressed_len = grr->pos;
+      gfi->free_compressed = Gif_DeleteArrayFunc;
+      grr->v = 0;
+      grr->cap = 0;
+    } else
+      gfi->compressed = 0;
+  }
+  grr->pos = 0;
 }
 
-
-static int get_color_table_size(Gif_Stream *, Gif_Image *, Gif_Writer *);
-
 int
-Gif_FullCompressImage(Gif_Stream *gfs, Gif_Image *gfi, int flags)
+Gif_FullCompressImage(Gif_Stream *gfs, Gif_Image *gfi,
+		      const Gif_CompressInfo *gcinfo)
 {
   int ok = 0;
   uint8_t min_code_bits;
   Gif_Writer grr;
-  Gif_Context gfc;
+  Gif_CodeTable gfc;
 
-  if (gfi->compressed && gfi->free_compressed) {
-    (*gfi->free_compressed)((void *)gfi->compressed);
-    gfi->compressed = 0;
-  }
+  gfc_init(&gfc);
 
-  gfc.nodes = Gif_NewArray(Gif_Node, NODES_SIZE);
-  gfc.links = Gif_NewArray(Gif_Node *, LINKS_SIZE);
-
-  grr.v = Gif_NewArray(uint8_t, 1024);
-  grr.pos = 0;
-  grr.cap = 1024;
+  grr.v = NULL;
+  grr.pos = grr.cap = 0;
   grr.byte_putter = memory_byte_putter;
   grr.block_putter = memory_block_putter;
-  grr.flags = flags;
+  if (gcinfo)
+    grr.gcinfo = *gcinfo;
+  else
+    Gif_InitCompressInfo(&grr.gcinfo);
   grr.global_size = get_color_table_size(gfs, 0, &grr);
   grr.local_size = get_color_table_size(gfs, gfi, &grr);
   grr.errors = 0;
 
-  if (!gfc.nodes || !gfc.links || !grr.v)
+  if (!gfc.nodes || !gfc.links) {
+    if (!(grr.gcinfo.flags & GIF_WRITE_SHRINK))
+      Gif_ReleaseCompressedImage(gfi);
     goto done;
+  }
 
-  min_code_bits = calculate_min_code_bits(gfs, gfi, &grr);
-  ok = write_image_data(gfi, min_code_bits, &gfc, &grr);
+  min_code_bits = calculate_min_code_bits(gfi, &grr);
+  ok = write_compressed_data(gfi, min_code_bits, &gfc, &grr);
+  save_compression_result(gfi, &grr, ok);
 
-  if ((flags & (GIF_WRITE_OPTIMIZE | GIF_WRITE_EAGER_CLEAR)) == GIF_WRITE_OPTIMIZE
+  if ((grr.gcinfo.flags & (GIF_WRITE_OPTIMIZE | GIF_WRITE_EAGER_CLEAR))
+      == GIF_WRITE_OPTIMIZE
       && grr.cleared && ok) {
-    uint8_t *old_v = grr.v;
-    uint32_t old_pos = grr.pos;
-    grr.v = Gif_NewArray(uint8_t, grr.cap);
-    grr.pos = 0;
-    grr.flags = flags | GIF_WRITE_EAGER_CLEAR;
-    if (grr.v && write_image_data(gfi, min_code_bits, &gfc, &grr)
-	&& grr.pos < old_pos) {
-      Gif_DeleteArray(old_v);
-      goto done;
-    }
-    Gif_DeleteArray(grr.v);
-    grr.v = old_v;
-    grr.pos = old_pos;
+    grr.gcinfo.flags |= GIF_WRITE_EAGER_CLEAR | GIF_WRITE_SHRINK;
+    if (write_compressed_data(gfi, min_code_bits, &gfc, &grr))
+      save_compression_result(gfi, &grr, 1);
   }
 
  done:
-  if (!ok) {
-    Gif_DeleteArray(grr.v);
-    grr.v = 0;
-  }
-  gfi->compressed = grr.v;
-  gfi->compressed_len = grr.pos;
-  gfi->free_compressed = Gif_DeleteArrayFunc;
+  Gif_DeleteArray(grr.v);
   Gif_DeleteArray(gfc.nodes);
   Gif_DeleteArray(gfc.links);
   return grr.v != 0;
@@ -500,7 +568,7 @@ Gif_FullCompressImage(Gif_Stream *gfs, Gif_Image *gfi, int flags)
 
 
 static int
-get_color_table_size(Gif_Stream *gfs, Gif_Image *gfi, Gif_Writer *grr)
+get_color_table_size(const Gif_Stream *gfs, Gif_Image *gfi, Gif_Writer *grr)
 {
   Gif_Colormap *gfcm = (gfi ? gfi->local : gfs->global);
   int ncol, totalcol, i;
@@ -513,7 +581,7 @@ get_color_table_size(Gif_Stream *gfs, Gif_Image *gfi, Gif_Writer *grr)
 
   /* Possibly bump up 'ncol' based on 'transparent' values, if
      careful_min_code_bits */
-  if (grr->flags & GIF_WRITE_CAREFUL_MIN_CODE_SIZE) {
+  if (grr->gcinfo.flags & GIF_WRITE_CAREFUL_MIN_CODE_SIZE) {
     if (gfi && gfi->transparent >= ncol)
       ncol = gfi->transparent + 1;
     else if (!gfi)
@@ -554,7 +622,8 @@ write_color_table(Gif_Colormap *gfcm, int totalcol, Gif_Writer *grr)
 
 
 static int
-write_image(Gif_Stream *gfs, Gif_Image *gfi, Gif_Context *gfc, Gif_Writer *grr)
+write_image(Gif_Stream *gfs, Gif_Image *gfi, Gif_CodeTable *gfc,
+            Gif_Writer *grr)
 {
   uint8_t min_code_bits, packed = 0;
   grr->local_size = get_color_table_size(gfs, gfi, grr);
@@ -579,13 +648,15 @@ write_image(Gif_Stream *gfs, Gif_Image *gfi, Gif_Context *gfc, Gif_Writer *grr)
     write_color_table(gfi->local, grr->local_size, grr);
 
   /* calculate min_code_bits here (because calculation may involve
-     recompression, if GIF_WRITE_CAREFUL_MIN_CODE_BITS is true) */
-  min_code_bits = calculate_min_code_bits(gfs, gfi, grr);
+     recompression, if GIF_WRITE_CAREFUL_MIN_CODE_SIZE is true) */
+  min_code_bits = calculate_min_code_bits(gfi, grr);
 
   /* use existing compressed data if it exists. This will tend to whip
      people's asses who uncompress an image, keep the compressed data around,
      but modify the uncompressed data anyway. That sucks. */
-  if (gfi->compressed) {
+  if (gfi->compressed
+      && (!(grr->gcinfo.flags & GIF_WRITE_CAREFUL_MIN_CODE_SIZE)
+          || gfi->compressed[0] == min_code_bits)) {
     uint8_t *compressed = gfi->compressed;
     uint32_t compressed_len = gfi->compressed_len;
     while (compressed_len > 0) {
@@ -596,7 +667,7 @@ write_image(Gif_Stream *gfs, Gif_Image *gfi, Gif_Context *gfc, Gif_Writer *grr)
     }
 
   } else
-    write_image_data(gfi, min_code_bits, gfc, grr);
+    write_compressed_data(gfi, min_code_bits, gfc, grr);
 
   return 1;
 }
@@ -653,7 +724,7 @@ write_graphic_control_extension(Gif_Image *gfi, Gif_Writer *grr)
 
 
 static void
-blast_data(uint8_t *data, int len, Gif_Writer *grr)
+blast_data(const uint8_t *data, int len, Gif_Writer *grr)
 {
   while (len > 0) {
     int s = len > 255 ? 255 : len;
@@ -682,7 +753,7 @@ write_comment_extensions(Gif_Comment *gfcom, Gif_Writer *grr)
   for (i = 0; i < gfcom->count; i++) {
     gifputbyte('!', grr);
     gifputbyte(0xFE, grr);
-    blast_data((uint8_t *)gfcom->str[i], gfcom->len[i], grr);
+    blast_data((const uint8_t *)gfcom->str[i], gfcom->len[i], grr);
   }
 }
 
@@ -690,7 +761,7 @@ write_comment_extensions(Gif_Comment *gfcom, Gif_Writer *grr)
 static void
 write_netscape_loop_extension(uint16_t value, Gif_Writer *grr)
 {
-  gifputblock((uint8_t *)"!\xFF\x0BNETSCAPE2.0\x03\x01", 16, grr);
+  gifputblock((const uint8_t *)"!\xFF\x0BNETSCAPE2.0\x03\x01", 16, grr);
   gifputunsigned(value, grr);
   gifputbyte(0, grr);
 }
@@ -708,7 +779,7 @@ write_generic_extension(Gif_Extension *gfex, Gif_Writer *grr)
     int len = gfex->application ? strlen(gfex->application) : 0;
     if (len) {
       gifputbyte(len, grr);
-      gifputblock((uint8_t *)gfex->application, len, grr);
+      gifputblock((const uint8_t *)gfex->application, len, grr);
     }
   }
   while (pos + 255 < gfex->length) {
@@ -732,10 +803,9 @@ write_gif(Gif_Stream *gfs, Gif_Writer *grr)
   int i;
   Gif_Image *gfi;
   Gif_Extension *gfex = gfs->extensions;
-  Gif_Context gfc;
+  Gif_CodeTable gfc;
 
-  gfc.nodes = Gif_NewArray(Gif_Node, NODES_SIZE);
-  gfc.links = Gif_NewArray(Gif_Node *, LINKS_SIZE);
+  gfc_init(&gfc);
   if (!gfc.nodes || !gfc.links)
     goto done;
 
@@ -750,9 +820,9 @@ write_gif(Gif_Stream *gfs, Gif_Writer *grr)
 	isgif89a = 1;
     }
     if (isgif89a)
-      gifputblock((uint8_t *)"GIF89a", 6, grr);
+      gifputblock((const uint8_t *)"GIF89a", 6, grr);
     else
-      gifputblock((uint8_t *)"GIF87a", 6, grr);
+      gifputblock((const uint8_t *)"GIF87a", 6, grr);
   }
 
   write_logical_screen_descriptor(gfs, grr);
@@ -794,13 +864,17 @@ write_gif(Gif_Stream *gfs, Gif_Writer *grr)
 
 
 int
-Gif_FullWriteFile(Gif_Stream *gfs, int flags, FILE *f)
+Gif_FullWriteFile(Gif_Stream *gfs, const Gif_CompressInfo *gcinfo,
+		  FILE *f)
 {
   Gif_Writer grr;
   grr.f = f;
   grr.byte_putter = file_byte_putter;
   grr.block_putter = file_block_putter;
-  grr.flags = flags;
+  if (gcinfo)
+    grr.gcinfo = *gcinfo;
+  else
+    Gif_InitCompressInfo(&grr.gcinfo);
   grr.errors = 0;
   return write_gif(gfs, &grr);
 }
