@@ -1,9 +1,9 @@
 /* quantize.c - Histograms and quantization for gifsicle.
-   Copyright (C) 1997-2011 Eddie Kohler, ekohler@gmail.com
+   Copyright (C) 1997-2013 Eddie Kohler, ekohler@gmail.com
    This file is part of gifsicle.
 
    Gifsicle is free software. It is distributed under the GNU Public License,
-   version 2 or later; you can copy, distribute, or alter it at will, as long
+   version 2; you can copy, distribute, or alter it at will, as long
    as this notice is kept intact and this source code is made available. There
    is no warranty, express or implied. */
 
@@ -970,7 +970,7 @@ colormap_stream(Gif_Stream *gfs, Gif_Colormap *new_cm,
       Gif_SetUncompressedImage(gfi, new_data, Gif_DeleteArrayFunc, 0);
 
       if (only_compressed) {
-	Gif_FullCompressImage(gfs, gfi, gif_write_flags);
+	Gif_FullCompressImage(gfs, gfi, &gif_write_info);
 	Gif_ReleaseUncompressedImage(gfi);
       }
 
@@ -1014,6 +1014,8 @@ colormap_stream(Gif_Stream *gfs, Gif_Colormap *new_cm,
      only that subset, just as if we'd piped the output of 'gifsicle
      --use-colormap=X' through 'gifsicle' another time. */
   gfs->global = Gif_CopyColormap(new_cm);
+  for (j = 0; j < new_cm->ncol; ++j)
+      gfs->global->col[j].haspixel = 0;
   if (compress_new_cm) {
     /* only bother to recompress if we'll get anything out of it */
     compress_new_cm = 0;
@@ -1061,7 +1063,7 @@ colormap_stream(Gif_Stream *gfs, Gif_Colormap *new_cm,
 	gfi->transparent = map[gfi->transparent];
 
       if (only_compressed) {
-	Gif_FullCompressImage(gfs, gfi, gif_write_flags);
+	Gif_FullCompressImage(gfs, gfi, &gif_write_info);
 	Gif_ReleaseUncompressedImage(gfi);
       }
     }
