@@ -1,9 +1,9 @@
 /* support.c - Support functions for gifsicle.
-   Copyright (C) 1997-2011 Eddie Kohler, ekohler@gmail.com
+   Copyright (C) 1997-2013 Eddie Kohler, ekohler@gmail.com
    This file is part of gifsicle.
 
    Gifsicle is free software. It is distributed under the GNU Public License,
-   version 2 or later; you can copy, distribute, or alter it at will, as long
+   version 2; you can copy, distribute, or alter it at will, as long
    as this notice is kept intact and this source code is made available. There
    is no warranty, express or implied. */
 
@@ -29,31 +29,44 @@ verror(int need_file, int seriousness, const char *fmt, va_list val)
   char buffer[BUFSIZ];
   static char *printed_file = 0;
   static int just_printed_context = 0;
-  const char *iname = input_name ? input_name : "<stdin>", *prefix = "";
+  const char *initial_prefix = program_name;
+  const char *prefix = "";
+  const char *iname = input_name;
+  if (!iname)
+    iname = "<stdin>";
+  if (mode == EXPLODING && active_output_data.active_output_name)
+    iname = active_output_data.active_output_name;
 
   if (printed_file && strcmp(printed_file, iname) != 0) {
       free(printed_file);
       printed_file = 0;
   }
-  if (need_file && !printed_file) {
+  if (need_file == 2 && !printed_file)
+      initial_prefix = iname;
+  else if (need_file && !printed_file) {
       if (mode != BLANK_MODE && mode != MERGING && nested_mode != MERGING) {
 	  fprintf(stderr, "%s: While processing '%s':\n", program_name, iname);
 	  just_printed_context = 1;
 	  prefix = "  ";
+	  initial_prefix = "";
       }
       printed_file = malloc(strlen(iname) + 1);
       strcpy(printed_file, iname);
-  } else if (just_printed_context && seriousness == 0)
+  } else if (just_printed_context && seriousness == 0) {
       prefix = "  ";
-  else
+      initial_prefix = "";
+  } else
       just_printed_context = 0;
 
   if (seriousness > 2)
-      sprintf(pattern, "%s:%s fatal error: %%s\n", program_name, prefix);
+      sprintf(pattern, "%s%s%s fatal error: %%s\n",
+	      initial_prefix, *initial_prefix ? ":" : "", prefix);
   else if (seriousness == 1)
-      sprintf(pattern, "%s:%s warning: %%s\n", program_name, prefix);
+      sprintf(pattern, "%s%s%s warning: %%s\n",
+	      initial_prefix, *initial_prefix ? ":" : "", prefix);
   else
-      sprintf(pattern, "%s:%s %%s\n", program_name, prefix);
+      sprintf(pattern, "%s%s%s %%s\n",
+	      initial_prefix, *initial_prefix ? ":" : "", prefix);
 
   if (seriousness > 1)
     error_count++;
@@ -153,6 +166,7 @@ General options: Also --no-OPTION for info and verbose.\n\
                                 normal output is not suppressed.\n\
       --color-info, --cinfo     --info plus colormap details.\n\
       --extension-info, --xinfo --info plus extension details.\n\
+      --size-info, --sinfo      --info plus compression information.\n\
   -V, --verbose                 Prints progress information.\n\
   -h, --help                    Print this message and exit.\n\
       --version                 Print version number and exit.\n\
@@ -384,13 +398,13 @@ extension_info(FILE *where, Gif_Stream *gfs, Gif_Extension *gfex, int count)
 
 
 void
-stream_info(FILE *where, Gif_Stream *gfs, const char *filename,
-	    int colormaps, int extensions)
+stream_info(FILE *where, Gif_Stream *gfs, const char *filename, int flags)
 {
   Gif_Extension *gfex;
   int n;
 
-  if (!gfs) return;
+  if (!gfs)
+    return;
 
   verbose_endline();
   fprintf(where, "* %s %d image%s\n", (filename ? filename : "<stdin>"),
@@ -400,7 +414,8 @@ stream_info(FILE *where, Gif_Stream *gfs, const char *filename,
 
   if (gfs->global) {
     fprintf(where, "  global color table [%d]\n", gfs->global->ncol);
-    if (colormaps) colormap_info(where, gfs->global, "  |");
+    if (flags & INFO_COLORMAPS)
+      colormap_info(where, gfs->global, "  |");
     fprintf(where, "  background %d\n", gfs->background);
   }
 
@@ -413,9 +428,9 @@ stream_info(FILE *where, Gif_Stream *gfs, const char *filename,
     fprintf(where, "  loop count %u\n", (unsigned)gfs->loopcount);
 
   for (n = 0, gfex = gfs->extensions; gfex; gfex = gfex->next, n++)
-    if (extensions)
+    if (flags & INFO_EXTENSIONS)
       extension_info(where, gfs, gfex, n);
-  if (n && !extensions)
+  if (n && !(flags & INFO_EXTENSIONS))
     fprintf(where, "  extensions %d\n", n);
 }
 
@@ -425,10 +440,11 @@ static char *disposal_names[] = {
 };
 
 void
-image_info(FILE *where, Gif_Stream *gfs, Gif_Image *gfi, int colormaps)
+image_info(FILE *where, Gif_Stream *gfs, Gif_Image *gfi, int flags)
 {
   int num;
-  if (!gfs || !gfi) return;
+  if (!gfs || !gfi)
+    return;
   num = Gif_ImageNumber(gfs, gfi);
 
   verbose_endline();
@@ -446,19 +462,18 @@ image_info(FILE *where, Gif_Stream *gfs, Gif_Image *gfi, int colormaps)
   if (gfi->transparent >= 0)
     fprintf(where, " transparent %d", gfi->transparent);
 
-#if defined(PRINT_SIZE)
-  if (gfi->compressed)
-    fprintf(where, " compressed size %u min_code_size %d", gfi->compressed_len, *gfi->compressed);
-#endif
-
   fprintf(where, "\n");
 
+  if ((flags & INFO_SIZES) && gfi->compressed)
+    fprintf(where, "    compressed size %u\n", gfi->compressed_len);
+
   if (gfi->comment)
     comment_info(where, gfi->comment, "    comment ");
 
   if (gfi->local) {
     fprintf(where, "    local color table [%d]\n", gfi->local->ncol);
-    if (colormaps) colormap_info(where, gfi->local, "    |");
+    if (flags & INFO_COLORMAPS)
+      colormap_info(where, gfi->local, "    |");
   }
 
   if (gfi->disposal || gfi->delay) {
@@ -835,7 +850,19 @@ read_text_colormap(FILE *f, const char *name)
       blue = (unsigned)(fblue + .5);
       goto found;
 
-    } else if (sscanf(buf, "#%2x%2x%2x", &red, &green, &blue) == 3) {
+    } else if (buf[0] == '#'
+               && strspn(buf + 1, "0123456789abcdefABCDEF") == 3
+               && (!buf[4] || isspace((unsigned char) buf[4]))) {
+      sscanf(buf + 1, "%1x%1x%1x", &red, &green, &blue);
+      red += red << 4;
+      green += green << 4;
+      blue += blue << 4;
+      goto found;
+
+    } else if (buf[0] == '#'
+               && strspn(buf + 1, "0123456789abcdefABCDEF") == 6
+               && (!buf[7] || isspace((unsigned char) buf[7]))) {
+      sscanf(buf + 1, "%2x%2x%2x", &red, &green, &blue);
      found:
       if (red > 255) red = 255;
       if (green > 255) green = 255;
@@ -1069,7 +1096,7 @@ find_color_or_error(Gif_Color *color, Gif_Stream *gfs, Gif_Image *gfi,
 
   index = Gif_FindColor(gfcm, color);
   if (index < 0 && color_context)
-    error(0, "%s color not in colormap", color_context);
+    error(2, "%s color not in colormap", color_context);
   return index;
 }
 
@@ -1376,6 +1403,22 @@ analyze_crop(int nmerger, Gt_Crop *crop, int compress_immediately)
 }
 
 
+static inline int
+apply_frame_transparent(Gif_Image *gfi, Gt_Frame *fr)
+{
+    int old_transparent = gfi->transparent;
+    if (fr->transparent.haspixel == 255)
+        gfi->transparent = -1;
+    else if (fr->transparent.haspixel) {
+        gfi->transparent =
+            find_color_or_error(&fr->transparent, fr->stream, gfi,
+                                "transparent");
+        if (gfi->transparent < 0)
+            fr->transparent.haspixel = 0;
+    }
+    return old_transparent;
+}
+
 Gif_Stream *
 merge_frame_interval(Gt_Frameset *fset, int f1, int f2,
 		     Gt_OutputData *output_data, int compress_immediately,
@@ -1446,6 +1489,14 @@ merge_frame_interval(Gt_Frameset *fset, int f1, int f2,
     if (merger[i]->crop && !merger[i]->crop->ready)
       analyze_crop(nmerger, merger[i]->crop, compress_immediately);
 
+  /* mark used colors */
+  for (i = 0; i < nmerger; ++i) {
+      int old_transp = apply_frame_transparent(merger[i]->image, merger[i]);
+      mark_used_colors(merger[i]->stream, merger[i]->image, merger[i]->crop,
+                       compress_immediately);
+      merger[i]->image->transparent = old_transp;
+  }
+
   /* copy stream-wide information from output_data */
   if (output_data->loopcount > -2)
     dest->loopcount = output_data->loopcount;
@@ -1456,7 +1507,7 @@ merge_frame_interval(Gt_Frameset *fset, int f1, int f2,
     Gt_Frame *fr = merger[i];
     Gif_Image *srci;
     Gif_Image *desti;
-    int old_transparent;
+    int old_transp;
 
     /* First, check for extensions */
     {
@@ -1504,12 +1555,7 @@ merge_frame_interval(Gt_Frameset *fset, int f1, int f2,
 
     /* It was pretty stupid to remove this code, which I did between 1.2b6 and
        1.2 */
-    old_transparent = srci->transparent;
-    if (fr->transparent.haspixel == 255)
-      srci->transparent = -1;
-    else if (fr->transparent.haspixel)
-      srci->transparent =
-	find_color_or_error(&fr->transparent, fr->stream, srci, "transparent");
+    old_transp = apply_frame_transparent(srci, fr);
 
     /* Is it ok to use the old image's compressed version? */
     /* 11.Feb.2003 -- It is NOT ok to use the old image's compressed version
@@ -1522,7 +1568,7 @@ merge_frame_interval(Gt_Frameset *fset, int f1, int f2,
 
     desti = merge_image(dest, fr->stream, srci, same_compressed_ok);
 
-    srci->transparent = old_transparent; /* restore real transparent value */
+    srci->transparent = old_transp; /* restore real transparent value */
 
     /* Flipping and rotating, and also setting the screen size */
     if (fr->flip_horizontal || fr->flip_vertical || fr->rotation)
@@ -1563,7 +1609,7 @@ merge_frame_interval(Gt_Frameset *fset, int f1, int f2,
     /* compress immediately if possible to save on memory */
     if (desti->img) {
       if (compress_immediately > 0) {
-	Gif_FullCompressImage(dest, desti, gif_write_flags);
+	Gif_FullCompressImage(dest, desti, &gif_write_info);
 	Gif_ReleaseUncompressedImage(desti);
       } else if (desti->compressed)
 	Gif_ReleaseCompressedImage(desti);
